# --------------------------------------------------------------------
# agent.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Thursday February 16, 2023
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import asyncio
import random
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import StrEnum, auto
from typing import Any, Coroutine, Iterable, Optional, Union

import waterlog
from commandmap import CommandMap

from bivalve.aio import Connection, Server, Stream, StreamConnection
from bivalve.call import Call, Response
from bivalve.datatypes import ArgV
from bivalve.util import async_wrap, get_millis, is_iterable

log = waterlog.get(__name__)


# --------------------------------------------------------------------
class Role(StrEnum):
    NONE = auto()
    PEER = auto()
    DOWNSTREAM = auto()
    UPSTREAM = auto()


# --------------------------------------------------------------------
@dataclass
class ConnectionContext:
    conn: Connection
    role: Role
    task: asyncio.Task
    ack_ttl: Optional[datetime] = None
    syn_at: datetime = datetime.min
    call_map: dict[int, Call] = field(default_factory=dict)


# --------------------------------------------------------------------
class BivalveAgent:
    def __init__(
        self,
        max_peers=0,  # no maximum connections
        syn_schedule=timedelta(seconds=10),
        syn_timeout=timedelta(seconds=5),
        syn_jitter=5,
        loop_duration_ms=150,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        incoming_role: Role = Role.DOWNSTREAM,
        outgoing_role: Role = Role.UPSTREAM,
    ):
        self._commands = CommandMap(self)
        self._conn_ctx_map: dict[int, ConnectionContext] = {}
        self._functions = CommandMap(self, prefix="fn_")
        self._loop = loop
        self._max_peers = max_peers
        self._servers: list[Server] = []
        self._shutdown_event = asyncio.Event()
        self._loop_duration_ms = loop_duration_ms
        self._syn_jitter = syn_jitter
        self._syn_schedule = syn_schedule
        self._syn_timeout = syn_timeout
        self._scheduled_tasks: set[asyncio.Task] = set()
        self._incoming_role = incoming_role
        self._outgoing_role = outgoing_role

    @property
    def running(self) -> bool:
        return bool(self._conn_ctx_map or self._servers)

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if self._loop is None:
            self._loop = asyncio.get_event_loop()
        return self._loop

    def connection_contexts(self):
        yield from self._conn_ctx_map.values()

    def peers(self) -> Iterable[Connection]:
        for ctx in self.connection_contexts():
            yield ctx.conn

    def upstream_peers(self) -> Iterable[Connection]:
        for ctx in self.connection_contexts():
            if ctx.role in (Role.PEER, Role.UPSTREAM):
                yield ctx.conn

    def downstream_peers(self) -> Iterable[Connection]:
        for ctx in self.connection_contexts():
            if ctx.role in (Role.PEER, Role.DOWNSTREAM):
                yield ctx.conn

    async def serve(self, **kwargs) -> Server:
        self._check_max_peers()

        server = await Server.serve(self.on_incoming_stream, **kwargs)
        log.info(f"Serving peers on {server}.")
        self._servers.append(server)
        return server

    async def connect(self, role=Role.NONE, **kwargs) -> Connection:
        self._check_max_peers()

        if role == Role.NONE:
            role = self._outgoing_role

        stream = await Stream.connect(**kwargs)
        conn = StreamConnection(stream)
        self.add_connection(conn, role)

        return conn

    def _check_max_peers(self):
        if self._max_peers and len(self._conn_ctx_map) >= self._max_peers:
            log.warning(
                f"Cancelled peer connection: maximum number of peers reached ({self._max_peers})."
            )
            raise RuntimeError("Maximum number of peer connections reached.")

    def bridge(self, role=Role.NONE) -> Connection:
        self._check_max_peers()

        send_queue: asyncio.Queue[ArgV] = asyncio.Queue()
        recv_queue: asyncio.Queue[ArgV] = asyncio.Queue()

        if role == Role.NONE:
            role = self._incoming_role

        our_conn = Connection.bridge(send_queue, recv_queue)
        their_conn = Connection.bridge(recv_queue, send_queue)
        log.info(f"Bridge connected on {our_conn}.")
        self.add_connection(our_conn, role)
        return their_conn

    async def on_incoming_stream(self, stream: Stream):
        if self._max_peers and len(self._conn_ctx_map) >= self._max_peers:
            await stream.close()
            log.warning(
                f"Rejected incoming connection on {stream}: maximum number of peers reached ({self._max_peers})."
            )
            return

        conn = StreamConnection(stream)
        log.info(f"Incoming peer connected: {stream}")
        self.add_connection(conn, self._incoming_role)

    def add_connection(self, conn: Connection, role=Role.UPSTREAM):
        self._conn_ctx_map[conn.id] = ConnectionContext(
            conn, role, asyncio.create_task(self.communicate(conn))
        )
        self.schedule(self._on_connect(conn))

    def schedule(self, awaitable: Coroutine) -> asyncio.Task:
        """
        Schedule a task to be run in parallel.
        """
        task: asyncio.Task = self.loop.create_task(awaitable)
        self._scheduled_tasks.add(task)
        task.add_done_callback(self._scheduled_tasks.discard)
        return task

    async def _on_connect(self, conn: Connection):
        try:
            await async_wrap(self.on_connect, conn)
        except Exception:
            log.exception("Error occurred during `on_connect()` handler.")

    def on_connect(self, conn: Connection):
        pass

    async def _on_disconnect(self, conn: Connection):
        try:
            await async_wrap(self.on_disconnect, conn)
        except Exception:
            log.exception("Error occurred during `on_disconnect()` handler.")

    def on_disconnect(self, conn: Connection):
        pass

    async def _on_unrecognized_command(self, conn: Connection, *argv):
        try:
            await async_wrap(self.on_unrecognized_command, conn, *argv)
        except Exception:
            log.exception("Error occurred during `on_unrecognized_command()` handler.")

    async def _on_unrecognized_function(self, conn: Connection, fn_name: str, *argv):
        return await async_wrap(self.on_unrecognized_function, conn, fn_name, *argv)

    def on_unrecognized_command(self, conn: Connection, *argv):
        pass

    def on_unrecognized_function(self, conn: Connection, fn_name: str, *argv):
        raise NotImplementedError()

    async def _on_startup(self):
        try:
            await async_wrap(self.on_startup)
        except Exception:
            log.exception("Error occurred during `on_startup()` handler.")

    def on_startup(self):
        pass

    async def _on_shutdown(self):
        try:
            await async_wrap(self.on_shutdown)
        except Exception:
            log.exception("Error occurred during `on_shutdown()` handler.")

    def on_shutdown(self):
        pass

    def disconnect(self, conn: Connection, notify=True):
        ctx = self._conn_ctx_map.get(conn.id)
        if ctx and ctx.task:
            ctx.task.cancel()
        if ctx:
            ctx.task = asyncio.create_task(self._cleanup(conn, notify))

    async def _cleanup(self, conn: Connection, notify=True):
        if notify:
            await conn.try_send("bye")

        await conn.close()

        if conn.id in self._conn_ctx_map:
            del self._conn_ctx_map[conn.id]
            log.info(f"Peer disconnected: {conn}")
            self.schedule(self._on_disconnect(conn))

    async def maintain(self):
        trash: list[Connection] = []

        now = datetime.now()

        for ctx in self._conn_ctx_map.values():
            try:
                if ctx.ack_ttl and ctx.ack_ttl <= now:
                    log.warning(f"Peer keepalive timed out: {ctx.conn}")
                    trash.append(ctx.conn)
                elif ctx.syn_at <= now:
                    await ctx.conn.send("syn")
                    ctx.syn_at = datetime.max
                    ctx.ack_ttl = now + self._syn_timeout
                    self._cleanup_calls(now, ctx)

            except Exception:
                trash.append(ctx.conn)
                log.exception(
                    f"Error managing connection for peer, closing connection: {ctx.conn}"
                )

        for conn in trash:
            await self._cleanup(conn, notify=False)

        if self._shutdown_event.is_set():
            await self._shutdown()

    def _cleanup_calls(self, now: datetime, ctx: ConnectionContext):
        trash: list[int] = []

        for call in ctx.call_map.values():
            if now >= call.expires_at:
                if not call.future.done():
                    call.future.set_exception(TimeoutError())
                trash.append(call.id)

        for call_id in trash:
            del ctx.call_map[call_id]

    async def process_command(self, conn: Connection, *argv: str):
        try:
            if len(argv) < 1:
                raise ValueError("No peer command was specified.")
            if argv[0] not in self._commands:
                raise ValueError(f"Peer command is not recognized: {argv[0]}.")
            command = self._commands[argv[0]]
            self.schedule(command(conn, *argv[1:]))

        except ConnectionError:
            raise

        except Exception:
            log.exception("Error processing peer command.")

    async def communicate(self, conn: Connection):
        while await conn.alive():
            try:
                argv = await conn.recv()
                try:
                    await self.process_command(conn, *argv)
                except ValueError:
                    log.exception("Received an unrecognized peer command.")
                    self.schedule(self._on_unrecognized_command(conn, *argv))
            except ConnectionAbortedError:
                await self._cleanup(conn, notify=False)
            except ConnectionError:
                if await conn.alive():
                    log.exception(f"Connection lost with peer {conn.id}.")
                await self._cleanup(conn, notify=False)

    async def run(self):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        self.schedule(self._on_startup())

        while self.running:
            now_ms = get_millis()
            await self.maintain()
            sleep_timeout = max(0, self._loop_duration_ms - (get_millis() - now_ms))
            await asyncio.sleep(sleep_timeout / 1000)

    def shutdown(self):
        self._shutdown_event.set()

    async def _shutdown(self):
        if not self.running:
            return

        log.info("Shutting down.")

        await self._on_shutdown()

        try:
            for ctx in list(self._conn_ctx_map.values()):
                await self._cleanup(ctx.conn)
            for server in self._servers:
                server.close()

        except Exception:
            log.exception("Error while shutting down.")

        finally:
            self._conn_ctx_map.clear()
            self._servers.clear()
            log.info("Shutdown complete.")

    async def send(self, *argv):
        await self.send_to(self.downstream_peers(), *argv)

    async def send_to(self, peer: Union[Connection, Iterable[Connection]], *argv):
        if is_iterable(peer):
            assert isinstance(peer, Iterable)
            await asyncio.gather(*(conn.send(*argv) for conn in peer))
        else:
            assert isinstance(peer, Connection)
            await peer.send(*argv)

    def call(
        self,
        *params: Any,
        timeout_ms: int = 10000,
    ) -> Call:
        try:
            conn_id = random.choice([*self.upstream_peers()]).id
        except IndexError:
            raise RuntimeError("No upstream peers to call.")

        ctx = self._conn_ctx_map.get(conn_id)
        if ctx is None:
            raise ValueError("Not a connected peer: id={conn.id}.")

        fn, *argv = [str(x) for x in params]
        call = Call(fn, argv)
        if timeout_ms == 0:
            call.expires_at = datetime.max
        else:
            call.expires_at = datetime.now() + timedelta(milliseconds=timeout_ms)
        ctx.call_map[call.id] = call
        self.schedule(ctx.conn.send(*call.to_call_cmd_argv()))
        return call

    async def cmd_syn(self, conn: Connection):
        await conn.send("ack")

    async def cmd_ack(self, conn: Connection):
        ctx = self._conn_ctx_map[conn.id]
        ctx.ack_ttl = None
        ctx.syn_at = (
            datetime.now()
            + self._syn_schedule
            + timedelta(seconds=random.randint(-self._syn_jitter, self._syn_jitter))
        )

    async def cmd_bye(self, conn: Connection):
        self.disconnect(conn, notify=False)

    async def cmd_call(self, conn: Connection, call_id: str, fn_name: str, *argv):
        if fn_name in self._functions:
            function = self._functions[fn_name]

        else:

            async def wrapper(conn: Connection, *argv):
                return await self._on_unrecognized_function(conn, fn_name, *argv)

            function = wrapper

        try:
            result = await async_wrap(function, conn, *argv)

        except NotImplementedError:
            log.error(
                f"Received call for an undefined function `{fn_name}` id={call_id}"
            )
            await conn.send(
                "return",
                call_id,
                Response.Code.ERROR,
                Response.Errors.UNDEFINED_FUNCTION,
            )
            return

        except Exception as e:
            log.exception(f"Error processing peer call to function `{fn_name}`.")
            await conn.send(
                "return",
                call_id,
                Response.Code.ERROR,
                Response.Errors.RUNTIME_ERROR,
                str(e),
            )
            return

        try:
            if is_iterable(result):
                await conn.send("return", call_id, Response.Code.OK, *result)
            else:
                await conn.send("return", call_id, Response.Code.OK, result)

        except Exception:
            log.exception("Failed to send result back to caller.")

    async def cmd_return(
        self, conn: Connection, call_id_str: str, response_code: str, *argv
    ):
        ctx = self._conn_ctx_map[conn.id]

        call_id = int(call_id_str)
        call = ctx.call_map[call_id]
        call.future.set_result(Response(Response.Code(response_code), [*argv]))
        del ctx.call_map[call_id]



# --------------------------------------------------------------------
# util.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Saturday February 11, 2023
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import asyncio
import inspect
import json
import time
from typing import Any, Sequence


# --------------------------------------------------------------------
class Borg:
    _shared_state: dict[str, Any] = {}

    def __init__(self):
        self.__dict__ = self._shared_state


# --------------------------------------------------------------------
async def async_wrap(f, *args, **kwargs):
    """
    From xeno.utils

    Wraps a normal function in a coroutine.  If the given function
    is already a coroutine function, we simply await it.
    """

    if not asyncio.iscoroutinefunction(f):
        return f(*args, **kwargs)
    return await f(*args, **kwargs)


# --------------------------------------------------------------------
def is_iterable(obj: Any) -> bool:
    """Determine if the given object is an iterable sequence other than a string or byte array."""
    return (
        isinstance(obj, Sequence)
        and not isinstance(obj, (str, bytes, bytearray))
        or inspect.isgenerator(obj)
    )


# --------------------------------------------------------------------
def str_escape(s: str) -> str:
    """
    Use the JSON library to escape a string for use as a literal.
    """

    return json.dumps(s).strip('"')


# --------------------------------------------------------------------
def get_millis() -> int:
    """
    Get the number of milliseconds since the UNIX Epoch.
    Useful for performance timing.
    """
    return time.time_ns() // 1000000


# --------------------------------------------------------------------
def new_future() -> asyncio.Future:
    """
    Creates a future with the current event loop.
    """
    loop = asyncio.get_event_loop()
    return loop.create_future()



# --------------------------------------------------------------------
# bridge.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Saturday May 6, 2023
#
# Distributed under terms of the MIT license.
# --------------------------------------------------------------------

import asyncio
import logging
import sys

import waterlog

from .client import ExampleClient
from .server import ExampleServer


# --------------------------------------------------------------------
async def main():
    waterlog.setup()

    if "--debug" in sys.argv:
        waterlog.set_level(logging.DEBUG)

    server = ExampleServer()
    client = ExampleClient()
    client.add_connection(server.bridge())

    await asyncio.gather(server.run(), client.run())


# --------------------------------------------------------------------
if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python
# --------------------------------------------------------------------
# cli.py
#
# Author: Lain Musgrove (lain.musgrove@hearst.com)
# Date: Monday October 9, 2023
# --------------------------------------------------------------------

import argparse
import asyncio
import getpass
import os
import shlex
import signal
import ssl
import sys
import threading
import time
import traceback
from dataclasses import dataclass
from typing import Optional

import waterlog

from bivalve.agent import BivalveAgent
from bivalve.call import Call
from bivalve.nio import NonBlockingTextInput

# --------------------------------------------------------------------
log = waterlog.get(__name__)


# --------------------------------------------------------------------
@dataclass
class Config:
    host: Optional[str] = None
    port: Optional[int] = None
    sock: Optional[str] = None
    script = ""
    debug = False
    ssl = False
    verify_cert = True
    check_hostname = True
    args: Optional[list[str]] = None

    def _argparser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(description="bivalve agent client")

        parser.add_argument(
            "--host",
            "-H",
            type=str,
            help="Host to connect to.  Requires --port/-p to be specified.",
        )
        parser.add_argument(
            "--port",
            "-p",
            type=int,
            help="Port on the host to connect to.  Requires --host/-h to be specified.",
        )
        parser.add_argument(
            "--sock", "-s", type=str, help="Path to a UNIX socket file to connect to."
        )
        parser.add_argument(
            "--ssl",
            "-S",
            action="store_true",
            help="Connect to an SSL encrypted socket.",
        )
        parser.add_argument(
            "--debug", action="store_true", help="Print DEBUG level logs."
        )
        parser.add_argument(
            "--no-check-hostname",
            dest="check_hostname",
            action="store_false",
            help="Disable SSL cert hostname check.",
        )
        parser.add_argument(
            "--no-verify-cert",
            dest="verify_cert",
            action="store_false",
            help="Disable SSL cert verification.",
        )
        parser.add_argument(
            "--script",
            "-x",
            help="Execute semicolon-delimited commands from the given file, then exit.",
        )
        parser.add_argument(
            "args", nargs=argparse.REMAINDER, help="A command to execute."
        )

        return parser

    def parse_args(self):
        self._argparser().parse_args(namespace=self)
        if (self.host is None or self.port is None) and (self.sock is None):
            raise ValueError("One of host/port or sock must be specified.")
        return self


# --------------------------------------------------------------------
class ClientAgent(BivalveAgent):
    def __init__(self, config: Config):
        super().__init__()
        self.config = config

    def ctrlc_handler(self, *_):
        log.critical("Ctrl+C received.")
        self.shutdown()

    def _load_script(self, filename: str) -> list[str]:
        if filename == "-":
            script = sys.stdin.read()
        else:
            with open(filename, "r") as infile:
                script = infile.read()

        commands = script.split("\n")
        return commands

    async def run(self):
        signal.signal(signal.SIGINT, self.ctrlc_handler)
        loop = asyncio.get_event_loop()

        if self.config.args:
            commands = []
            if self.config.script:
                commands.extend(self._load_script(self.config.script))
            commands.append(shlex.join(self.config.args))
            thread = threading.Thread(target=self.script_thread, args=(loop, commands))

        elif self.config.script:
            commands = self._load_script(self.config.script)
            thread = threading.Thread(target=self.script_thread, args=(loop, commands))

        else:
            thread = threading.Thread(target=self.repl_thread, args=(loop,))

        ssl_ctx: Optional[ssl.SSLContext] = None

        if self.config.ssl:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            ssl_ctx.check_hostname = self.config.check_hostname
            ssl_ctx.verify_mode = ssl.CERT_NONE
            if self.config.verify_cert:
                ssl_ctx.verify_mode = ssl.CERT_REQUIRED

        await self.connect(
            host=self.config.host,
            port=self.config.port,
            path=self.config.sock,
            ssl=ssl_ctx,
        )

        thread.start()
        await super().run()
        thread.join()

    def on_disconnect(self, _):
        self.shutdown()

    def on_unrecognized_command(self, conn, *argv):
        print(f"<< CMD {' '.join(argv)}")

    def script_thread(self, loop, commands):
        current_call: Optional[Call] = None
        asyncio.set_event_loop(loop)
        quiet = False

        try:
            while not self._shutdown_event.is_set() and (commands or current_call):
                if current_call:
                    try:
                        result = current_call.future.result()
                        if not quiet and result.code == result.Code.OK:
                            sys.stdout.write(" ".join(result.content))
                            sys.stdout.flush()

                        elif result.code == result.Code.ERROR:
                            print(f"ERROR: {' '.join(result.content)}", file=sys.stderr)
                            break

                        current_call = None
                        quiet = False

                    except asyncio.InvalidStateError:
                        time.sleep(0.10)

                    except Exception:
                        traceback.print_exc()
                        current_call = None
                        quiet = False

                    continue

                command: str = commands.pop(0)
                command = command.format(**os.environ)

                argv = shlex.split(command)
                if not argv:
                    continue
                if argv[0] == "call":
                    current_call = self.call(*argv[1:], timeout_ms=0)
                elif argv[0] == "@call":
                    current_call = self.call(*argv[1:], timeout_ms=0)
                    quiet = True
                else:
                    asyncio.run_coroutine_threadsafe(
                        self.send_to(self.peers(), *argv), loop
                    )

        finally:
            self.shutdown()

    def repl_thread(self, loop):
        current_call: Optional[Call] = None
        asyncio.set_event_loop(loop)

        with NonBlockingTextInput() as ninput:
            while not self._shutdown_event.is_set():
                if current_call:
                    try:
                        result = current_call.future.result()

                        if result.code == result.Code.OK:
                            print()
                            print(f"<< OK {' '.join(result.content)}")

                        elif result.code == result.Code.ERROR:
                            print()
                            print(f"<< ERROR {' '.join(result.content)}")

                        current_call = None

                    except asyncio.InvalidStateError:
                        sys.stdout.write(".")
                        sys.stdout.flush()
                        time.sleep(0.25)

                    except Exception:
                        traceback.print_exc()
                        current_call = None
                    continue

                s = ninput.read("> ", timeout=0.25)
                if s is not None:
                    argv_in = shlex.split(s)
                    if not argv_in:
                        continue

                    argv = []

                    while argv_in:
                        arg = argv_in.pop(0)
                        if arg == "<>":
                            argv.append(getpass.getpass("pwd > "))
                        else:
                            argv.append(arg)

                    if argv[0] == "call":
                        current_call = self.call(*argv[1:], timeout_ms=0)

                    else:
                        asyncio.run_coroutine_threadsafe(
                            self.send_to(self.peers(), *argv), loop
                        )

        # Print a newline before we exit repl.
        print()


# --------------------------------------------------------------------
def main():
    waterlog.setup()
    config = Config().parse_args()

    if config.debug:
        waterlog.set_level("DEBUG")
    else:
        waterlog.set_level("WARNING")

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    client = ClientAgent(config)
    loop.run_until_complete(client.run())


# --------------------------------------------------------------------
if __name__ == "__main__":
    main()



# --------------------------------------------------------------------
# datatypes.py
#
# Author: Lain Musgrove (lain.musgrove@gmail.com)
# Date: Monday July 31, 2023
# --------------------------------------------------------------------

import asyncio
import threading
from dataclasses import dataclass
from typing import Any, Callable, Generic, Optional, TypeVar

# --------------------------------------------------------------------
T = TypeVar("T")
ArgV = list[str]
ArgVQueue = asyncio.Queue[ArgV]


# --------------------------------------------------------------------
class AtomicValue(Generic[T]):
    """
    A value with methods to support fetching and acting upon it atomically
    across multiple async coroutines.
    """

    def __init__(self, value: T):
        self.value = value
        self.lock = asyncio.Lock()

    async def __call__(self):
        async with self.lock:
            return self.value

    async def set(self, value: T):
        async with self.lock:
            self.value = value

    async def mutate(self, mutator: Callable[[T], T]):
        async with self.lock:
            self.value = mutator(self.value)


# --------------------------------------------------------------------
class AtomicBox(Generic[T]):
    """
    A single-item container with methods to support fetching and updating the
    stored contents atomically in an async context.
    """

    def __init__(self):
        self._value: AtomicValue[Optional[T]] = AtomicValue(None)

    async def is_empty(self) -> bool:
        value = await self._value()
        return value is None

    async def get(self) -> T:
        value = await self._value()
        if value is None:
            raise ValueError("No value has been supplied yet.")
        return value

    async def set(self, value: T):
        value = await self._value.set(value)

    async def clear(self):
        await self._value.set(None)


# --------------------------------------------------------------------
class ThreadAtomicCounter:
    """
    A thread-safe atomic auto-incrementing counter starting at 0,
    yielding numbers starting at 1 via `next()`.
    """

    def __init__(self):
        self.value = 0
        self.lock = threading.Lock()

    def next(self):
        with self.lock:
            self.value += 1
            return self.value


# --------------------------------------------------------------------
class PackedString:
    def __init__(self, value: Any, encoding="utf8"):
        self.bytes = str(value).encode(encoding)
        self.size = len(self.bytes)

    def to_bytes(self) -> bytes:
        return self.size.to_bytes(4) + self.bytes
